"""> Functions for registering tasks with myke."""

from __future__ import annotations

import collections.abc
import os
from functools import partial, wraps
from subprocess import CompletedProcess
from types import ModuleType
from typing import Any, Callable, Sequence

from .exceptions import NoTasksFoundError, TaskAlreadyRegisteredError
from .globals import ROOT_TASK_KEY, TASKS
from .run import sh
from .utils import _MykeSourceFileLoader, convert_to_command_string


def add_tasks(*args: Callable[..., Any], **kwargs: Callable[..., Any]) -> None:
    """Register the given callable(s) with myke.

    Arguments:
        *args: ...
        **kwargs: ...

    Raises:
        TaskAlreadyRegisteredError: ...

    Examples:
        >>> import myke
        ...
        >>> def say_hello(name):
        ...    print(f'Hello {name}.')
        ...
        >>> def say_goodbye(name):
        ...    print(f'Goodbye {name}.')
        ...
        >>> myke.add_tasks(say_hello, say_goodbye)
    """
    kwargs.update({x.__name__: x for x in args})

    kwargs = {
        (ROOT_TASK_KEY if k == ROOT_TASK_KEY else convert_to_command_string(k)): v
        for k, v in kwargs.items()
    }

    for k, v in kwargs.items():
        v_existing: Callable[..., Any] | None = TASKS.get(k, None)
        if v_existing:
            raise TaskAlreadyRegisteredError(
                (
                    f"Failed to import module '{v.__module__}': "
                    f"Task '{k}' already defined from module '{v_existing.__module__}'"
                ),
            )
        TASKS[k] = v


def import_mykefile(path: str) -> None:
    """Import tasks from another Mykefile.

    Args:
        path: path to the Mykefile

    Raises:
        NoTasksFoundError:

    Examples:
        >>> import myke
        ...
        >>> myke.import_mykefile('/path/to/tasks.py')  # doctest: +SKIP
    """
    n_tasks_before: int = len(TASKS)

    loader = _MykeSourceFileLoader(os.path.relpath(path), path)
    mod: ModuleType = ModuleType(loader.name)
    loader.exec_module(mod)

    if len(TASKS) <= n_tasks_before:
        raise NoTasksFoundError(path)


def import_module(name: str) -> None:
    """Import tasks from the given Python module.

    Args:
        name: name of the module.

    Raises:
        NoTasksFoundError:

    Examples:
        >>> import myke
        ...
        >>> myke.import_module('python_pkg.python_module')  # doctest: +SKIP
    """
    n_tasks_before: int = len(TASKS)

    __import__(name)

    if len(TASKS) <= n_tasks_before:
        raise NoTasksFoundError(name)


def task(
    func: Callable[..., Any] | None = None,
    *,
    name: str | None = None,
    root: bool | None = False,
) -> Callable[..., Any] | Callable[..., Callable[..., Any]]:
    """Function decorator to register functions with myke.

    Args:
        func: ...
        name: name of the command.
        root: if True, import this as the root command.

    Returns:
        ...

    Examples:
        >>> from myke import task
        ...
        >>> @task  # doctest: +SKIP
        ... def say_hello(name):
        ...    print(f'Hello {name}.')
        ...
        >>> @task  # doctest: +SKIP
        ... def say_goodbye(name):
        ...    print(f'Goodbye {name}.')
        ...
    """

    if not func:
        return partial(task, name=name, root=root)

    if root:
        name = ROOT_TASK_KEY

    if name:
        add_tasks(**{name: func})
    else:
        add_tasks(func)

    return func


def shell_task(
    func: Callable[..., str | Sequence[str]] | None = None,
    *,
    name: str | None = None,
    root: bool | None = False,
    capture_output: bool | None = False,
    echo: bool | None = True,
    check: bool | None = True,
    cwd: str | None = None,
    env: dict[str, str] | None = None,
    env_update: dict[str, str | None] | None = None,
    timeout: float | None = None,
    executable: str | None = None,
) -> (
    Callable[..., CompletedProcess[bytes | str]]
    | Callable[..., Callable[..., CompletedProcess[bytes | str]]]
):
    """Function decorator to register shell commands with myke.

    myke expects the function to return a string of one or more shell-commands,
    and will invoke the commands using `myke.run(..., shell=True)`.

    Args:
        func: ...
        name: name of the command.
        root: if True, import this as the root command.
        capture_output: ...
        echo: ...
        check: ...
        cwd: ...
        env: ...
        env_update: ...
        timeout: ...
        executable: ...

    Returns:
        ...

    Examples:
        >>> from myke import shell_task
        ...
        >>> @shell_task  # doctest: +SKIP
        ... def say_hello(name):
        ...    return f"echo 'Hello {name}.'"
        ...
        >>> @shell_task  # doctest: +SKIP
        ... def say_goodbye(name):
        ...    return f"echo 'Goodbye {name}.'"
        ...
    """
    if not func:
        return partial(
            shell_task,
            name=name,
            root=root,
            capture_output=capture_output,
            echo=echo,
            check=check,
            cwd=cwd,
            env=env,
            env_update=env_update,
            timeout=timeout,
            executable=executable,
        )

    @wraps(func)
    def _inner_func(*args: Any, **kwargs: Any) -> tuple[str | None, str | None, int]:
        assert func

        func_script: str = func(*args, **kwargs)

        if not isinstance(func_script, str) and not isinstance(
            func_script,
            collections.abc.Sequence,
        ):
            raise TypeError(
                "Expected a string to be returned for the `shell_task`",
            )

        return sh(
            func_script,
            capture_output=capture_output,
            echo=echo,
            check=check,
            cwd=cwd,
            env=env,
            env_update=env_update,
            timeout=timeout,
            executable=executable,
        )

    if not name:
        name = func.__name__

    return task(_inner_func, name=name, root=root)


# TODO: deprecate
task_sh = shell_task



"""> Functions for writing."""

import os
from pathlib import Path
from typing import Any, List, Optional, Union

from ..globals import DEFAULT_MYKEFILE
from ..utils import make_executable


class write:
    def __new__(  # type: ignore
        cls,
        content: Union[str, bytes],
        path: str,
        append: bool = False,
        overwrite: bool = False,
        **kwargs: Any,
    ) -> None:
        cls.text(
            content=content,
            path=path,
            append=append,
            overwrite=overwrite,
            **kwargs,
        )

    @staticmethod
    def text(
        content: Union[str, bytes],
        path: Union[str, Path],
        append: bool = False,
        overwrite: bool = False,
        encoding: str = "utf-8",
        **kwargs: Any,
    ) -> None:
        """Write text to a file.

        Args:
            content: ...
            path: ...
            append: ...
            overwrite: ...
            encoding: ...
            **kwargs: ...

        Raises:
            FileExistsError: if file exists and overwrite is False.

        Examples:
            >>> import myke
            ...
            >>> myke.write.text('hello world', '/path/to/file.txt')  # doctest: +SKIP
        """
        mode_default: str = "w"

        mode: str = kwargs.pop("mode", mode_default)

        if isinstance(path, str):
            path = Path(path)

        if path.exists():
            if overwrite:
                path.unlink()
            elif append and mode == mode_default:
                mode = "a"
            else:
                raise FileExistsError(path)

        if mode == mode_default:
            if isinstance(content, bytes):
                mode += "b"

            if not path.exists():
                mode += "+"

        with path.open(mode=mode, encoding=encoding, **kwargs) as f:
            f.write(content)

    @classmethod
    def lines(
        cls,
        content: List[Optional[str]],
        path: Union[str, Path],
        append: bool = False,
        overwrite: bool = False,
        **kwargs: Any,
    ) -> None:
        """Write lines of text to a file.

        Args:
            content: ...
            path: ...
            append: ...
            overwrite: ...
            **kwargs: ...

        Raises:
            FileExistsError: if file exists and overwrite is False.

        Examples:
            >>> import myke
            ...
            >>> myke.write.text(['hello', 'world'], '/path/to/file.txt')  # doctest: +SKIP
        """
        if isinstance(content, str):
            raise TypeError("expected a list of strings, not a string")
        cls.text(
            path=path,
            content=os.linesep.join([str(x) for x in content]),
            append=append,
            overwrite=overwrite,
            **kwargs,
        )

    @classmethod
    def mykefile(
        cls,
        path: Union[None, str, Path] = None,
    ) -> None:
        """Create a new Mykefile.

        Args:
            path: ...

        Raises:
            FileExistsError: if file exists and overwrite is False.


        Examples:
            >>> import myke
            ...
            >>> myke.write.mykefile('/path/to/Mykefile')  # doctest: +SKIP
        """
        if not path:
            path = Path.cwd()
        elif isinstance(path, str):
            path = Path(path)

        if path.is_dir():
            path = path / DEFAULT_MYKEFILE

        cls.text(
            path=path,
            content=r"""from myke import task, shell_task

@task(root=True)
def setup():
    # setup
    ...

    yield

    # teardown
    ...

@task
def do_work():
    ...

@shell_task
def more_work():
    return 'echo ...'
""",
        )

        make_executable(path)



"""> Functions that wrap around `subprocess.run` for added convenience."""

from __future__ import annotations

import json
import os
import subprocess
import sys
from functools import wraps
from typing import Any, Sequence

from .utils import split_and_trim_text

__all__ = [
    "run",
    "run_stdout",
    "run_stdout_lines",
    "sh",
    "sh_stdout",
    "sh_stdout_lines",
    "require",
]


def run(
    args: str | Sequence[str],
    capture_output: None | bool = False,
    echo: bool | None = True,
    check: bool | None = True,
    env: dict[str, str] | None = None,
    env_update: dict[str, str | None] | None = None,
    shell: bool | None = None,
    **kwargs: Any,
) -> subprocess.CompletedProcess[bytes | str]:
    r"""Thin wrapper around `subprocess.run`

    Args:
        args: ...
        capture_output: ...
        echo: ...
        check: ...
        env: ...
        env_update: ...
        shell: ...
        **kwargs: passed to `subprocess.run(...)`

    Returns:
        ...

    Examples:
        >>> import myke
        ...
        >>> myke.run(["python", "-c", "print('Hello World.')"])
        CompletedProcess(args=['python', '-c', "print('Hello World.')"], returncode=0)
        >>> myke.run("echo 'Hello World.'")
        CompletedProcess(args="echo 'Hello World.'", returncode=0)
        >>> p = myke.run("echo 'Hello World.'", capture_output=True, echo=False)
        >>> p.stdout
        b'Hello World.\n'
        >>> p = myke.run("exit 123", check=False)
        >>> p.returncode
        123
    """
    if shell is None:
        shell = isinstance(args, str) and " " in args

    env = env.copy() if env else os.environ.copy()

    if env_update:
        for k, v in env_update.items():
            if v is None:
                env.pop(k, None)
            else:
                env[k] = v

    if not echo and not capture_output:
        for k in ("stdout", "stderr"):
            kwargs[k] = subprocess.DEVNULL

    p: subprocess.CompletedProcess[str] = subprocess.run(
        args,
        shell=shell,
        env=env,
        capture_output=bool(capture_output),
        check=False,
        **kwargs,
    )

    try:
        if check:
            p.check_returncode()
    finally:
        if echo and capture_output:
            for x in (p.stdout, p.stderr):
                if x:
                    if isinstance(x, bytes):
                        x = x.decode()
                    print(x.rstrip(os.linesep))

    return p


@wraps(run)
def run_stdout(*args: Any, **kwargs: Any) -> str:
    """Shorthand for:

    `myke.run(..., capture_output=True, text=True, echo=False).stdout.strip()`

    Args:
        *args: ...
        **kwargs: ...

    Returns:
        ...

    Examples:
        >>> myke.run_stdout("echo 'Hello World.'")
        Hello World.
    """
    kwargs["capture_output"] = True
    kwargs["text"] = True
    kwargs["echo"] = kwargs.get("echo", False)
    p: subprocess.CompletedProcess[str] = run(*args, **kwargs)
    return p.stdout.strip()


@wraps(run)
def run_stdout_lines(*args: Any, **kwargs: Any) -> list[str]:
    """Similar to...

    `myke.run_stdout(...)`

    ...except that text is split on newlines and stripped of empty elements.

    Args:
        *args: ...
        **kwargs: ...

    Returns:
        ...

    Examples:
        >>> myke.run_stdout("echo '   Hello World.   ' '  ' '   Goodbye World.   '")
        ['Hello World.', 'Goodbye World.']
    """
    return split_and_trim_text(run_stdout(*args, **kwargs))


@wraps(run)
def sh(*args: Any, **kwargs: Any) -> subprocess.CompletedProcess[bytes | str]:
    """Shorthand for: `myke.run(..., shell=True)`

    Args:
        *args: ...
        **kwargs: ...

    Returns:
        ...
    """
    return run(*args, shell=True, **kwargs)


@wraps(sh)
def sh_stdout(*args: Any, **kwargs: Any) -> str:
    """Shorthand for: `myke.run_stdout(..., shell=True)`

    Args:
        *args: ...
        **kwargs: ...

    Returns:
        ...
    """
    return run_stdout(*args, shell=True, **kwargs)


@wraps(sh)
def sh_stdout_lines(*args: Any, **kwargs: Any) -> list[str]:
    """Shorthand for: `myke.run_stdout_lines(..., shell=True)`

    Args:
        *args: ...
        **kwargs: ...

    Returns:
        ...
    """
    return run_stdout_lines(*args, shell=True, **kwargs)


def _run_pip(
    *args: str,
    pip_args: list[str] | None = None,
    echo: bool = False,
    capture_output: bool = False,
    **kwargs: str,
) -> subprocess.CompletedProcess[str]:
    if not pip_args:
        pip_args = []

    return run(
        [
            sys.executable,
            "-m",
            "pip",
            "install",
            *pip_args,
            *args,
            *[f"{k}=={v}" for k, v in kwargs.items()],
        ],
        echo=echo,
        capture_output=capture_output,
        text=True,
    )


def require(
    *args: str,
    pip_args: list[str] | None = None,
    skip_check: bool = False,
    **kwargs: str,
) -> subprocess.CompletedProcess[str]:
    """Check for the given modules, and install them if they do not exist.

    Args:
        *args: modules to require.
        pip_args: args passed to `pip`.
        skip_check: don't check if module is already installed.
        **kwargs: modules to require.

    Returns:
        ...

    Examples:
        >>> import myke
        ...
        >>> myke.require('module-a==0.1.*', 'module-b==0.1.*', **{  # doctest: +SKIP
        ...     'module-c': '0.1.*',
        ... })
    """
    if not pip_args:
        pip_args = []

    for x in ["--disable-pip-version-check"]:
        if x not in pip_args:
            pip_args.append(x)

    dry_run_args: list[str] = (
        [] if skip_check else ["-qq", "--dry-run", "--report", "-"]
    )

    p: subprocess.CompletedProcess[str] = _run_pip(
        *args,
        pip_args=pip_args + dry_run_args,
        echo=False,
        capture_output=True,
        **kwargs,
    )

    if dry_run_args:
        if not p.stdout or p.returncode != 0:
            print(p.stdout, p.stderr, p.returncode)
            return p

        result: dict[str, Any] = json.loads(p.stdout)

        if result and result.get("install"):
            return _run_pip(
                *args,
                pip_args=pip_args,
                echo=True,
                capture_output=False,
                **kwargs,
            )

    return p



import os
import sys
from contextlib import suppress
from dataclasses import dataclass
from fnmatch import fnmatch
from inspect import getsource
from pathlib import Path
from subprocess import CalledProcessError
from typing import Any, Callable, List, Optional, Union

import yapx

from .__version__ import __version__
from .exceptions import NoTasksFoundError, TaskAlreadyRegisteredError
from .globals import DEFAULT_MYKEFILE, MYKE_VAR_NAME, ROOT_TASK_KEY, TASKS
from .io.echo import echo
from .io.write import write
from .tasks import import_module, import_mykefile
from .types import Annotated
from .utils import get_repo_root

__all__ = ["__version__", "main", "sys"]


def main(_file: Optional[Union[str, Path]] = None) -> None:
    @dataclass
    class MykeArgs(yapx.types.Dataclass):
        file: Annotated[
            Optional[List[Path]],
            yapx.arg(
                "myke-file",
                default=_file if _file else None,
                env="MYKE_FILE",
                group="myke parameters",
            ),
        ]
        module: Annotated[
            Optional[List[str]],
            yapx.arg(
                "myke-module",
                default=None,
                env="MYKE_MODULE",
                group="myke parameters",
            ),
        ]
        list_tasks: Annotated[
            Optional[bool],
            yapx.arg(
                "myke-tasks",
                default=None,
                group="myke parameters",
                exclusive=True,
            ),
        ]
        explain: Annotated[
            Optional[bool],
            yapx.arg(
                "myke-explain",
                default=None,
                group="myke parameters",
                exclusive=True,
            ),
        ]
        create: Annotated[
            Optional[bool],
            yapx.arg(
                "myke-create",
                default=None,
                group="myke parameters",
                exclusive=True,
            ),
        ]

    prog: str = str(_file) if _file else MYKE_VAR_NAME

    parser = yapx.ArgumentParser(
        prog=prog,
        prog_version=__version__,
        help_flags=["--myke-help"],
        version_flags=["--myke-version"],
        completion_flags=[],
        tui_flags=[],
    )
    parser.add_arguments(MykeArgs)

    args = sys.argv[1:]

    myke_args: MykeArgs
    task_args: List[str]
    myke_args, task_args = parser.parse_known_args_to_model(
        args,
        args_model=MykeArgs,
        skip_pydantic_validation=True,
    )
    assert isinstance(myke_args, MykeArgs)

    if _file:
        if not isinstance(_file, Path):
            _file = Path(_file)
        if not _file.exists():
            raise FileNotFoundError(_file)
        if not TASKS:
            raise NoTasksFoundError(_file)
        _file = _file.absolute()

    if myke_args.file:
        myke_args.file = [x.absolute() for x in myke_args.file]
    elif Path(DEFAULT_MYKEFILE).exists():
        myke_args.file = [Path(DEFAULT_MYKEFILE).absolute()]

    with suppress(FileNotFoundError):
        repo_root: Optional[Path] = get_repo_root()
        if repo_root:
            os.chdir(repo_root)

    if not myke_args.file:
        if Path(DEFAULT_MYKEFILE).exists():
            myke_args.file = [Path(DEFAULT_MYKEFILE).absolute()]
        else:
            myke_args.file = []

    if myke_args.create:
        out_file: Path = myke_args.file[0] if myke_args.file else Path(DEFAULT_MYKEFILE)
        write.mykefile(str(out_file))
        echo(f"Created: {out_file}")
        parser.exit()

    try:
        try:
            for f in myke_args.file:
                if f and (not _file or not f.samefile(_file)):
                    import_mykefile(str(f))
                    # TODO: os.environ["MYKE_FILE"] = str(myke_args.file)
        except FileNotFoundError as e:
            parser.print_help()
            echo(
                (
                    f"{os.linesep}"
                    f"'{e.filename}' not found. Create it using:"
                    f"{os.linesep}"
                    f"$ {prog} --myke-create --myke-file '{e.filename}'"
                    f"{os.linesep}"
                ),
            )
            parser.exit()

        if myke_args.module:
            for m in myke_args.module:
                import_module(m)
                # TODO: os.environ["MYKE_MODULE"] = x
    except TaskAlreadyRegisteredError as e:
        parser.error(str(e))

    root_task: Optional[Callable[..., Any]] = TASKS.pop(ROOT_TASK_KEY, None)

    if myke_args.explain:
        explain_this: Optional[Callable[..., Any]] = None

        if not task_args or task_args[0].startswith("-"):
            explain_this = root_task
            if explain_this is None:
                echo("There is no root task. Provide a task name to explain.")
        else:
            explain_this = TASKS.get(task_args[0], None)
            if explain_this is None:
                echo(f"Given task name not found: {task_args[0]}")

        if explain_this:
            echo(getsource(explain_this))

        parser.exit()

    if myke_args.list_tasks:
        echo.tasks(prog=prog)
        parser.exit()

    if task_args and not task_args[0].startswith("--"):
        for k in list(TASKS):
            if not fnmatch(k, task_args[0]):
                del TASKS[k]

    try:
        yapx.run(
            root_task,
            named_subcommands=TASKS,
            args=task_args,
            default_args=["--tui"],
            prog=prog,
            prog_version=__version__,
        )
    except CalledProcessError as e:
        print(e)
        if e.output:
            print(f"stdout: {e.output}")
        elif e.stderr:
            print(f"stderr: {e.stderr}")
        sys.exit(e.returncode)
    except KeyboardInterrupt:
        pass



"""> Functions for printing."""

import collections.abc
import os
from typing import Any, Dict, Iterable, List, Mapping, Optional, Sequence, Union

from ..globals import MYKE_VAR_NAME, TASKS


class echo:
    def __new__(cls, *args: Any, **kwargs: Any) -> None:  # type: ignore
        cls.text(*args, **kwargs)

    @staticmethod
    def _print(*args: Any, print_kwargs: Optional[Dict[str, Any]] = None) -> None:
        if not print_kwargs:
            print_kwargs = {}
        print(*args, **print_kwargs)

    @classmethod
    def text(cls, *args: Any, **kwargs: Any) -> None:
        """Prints text.

        Equivalent to base `print` command; only exists for consistency.

        Arguments:
            *args: ...
            **kwargs: ...

        Examples:
            >>> import myke
            ...
            >>> myke.echo.text('Hello World.')
            Hello World.
        """
        cls._print(*args, print_kwargs=kwargs)

    @classmethod
    def lines(
        cls,
        seq: Sequence[Optional[str]],
        linesep: str = os.linesep,
        print_kwargs: Optional[Dict[str, Any]] = None,
    ) -> None:
        """Prints lines of text.

        Arguments:
            seq: lines of text to print.
            linesep: line separator.
            print_kwargs: kwargs passed to `print`.

        Examples:
            >>> import myke
            ...
            >>> myke.echo.lines(['Hello World.', 'Goodbye World.'])
            Hello World.
            Goodbye World.
        """
        cls._print(linesep.join([str(x) for x in seq]), print_kwargs=print_kwargs)

    @classmethod
    def json(
        cls,
        obj: Any,
        print_kwargs: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> None:
        """Prints a dictionary as a JSON string.

        Arguments:
            obj: object to convert to JSON and print.
            print_kwargs: kwargs passed to `print`.
            **kwargs: kwargs passed to `json.dumps`.

        Examples:
            >>> import myke
            ...
            >>> myke.echo.json(
            ...     {'Messages': ['Hello World.', 'Goodbye World.']},
            ...     indent=4,
            ... )
            {
                "Messages": [
                    "Hello World.",
                    "Goodbye World."
                ]
            }
        """
        import json as _json

        cls._print(_json.dumps(obj, **kwargs), print_kwargs=print_kwargs)

    @classmethod
    def table(
        cls,
        obj: Union[Iterable[Iterable[Any]], Mapping[str, Iterable[Any]]],
        print_kwargs: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> None:
        """Prints a dictionary as a table.

        Arguments:
            obj: list of dicts print.
            print_kwargs: kwargs passed to `print`.
            **kwargs: kwargs passed to `tabulate`.

        Examples:
            >>> import myke
            ...
            >>> myke.echo.table([
            ...     {'row': 1, 'Message': 'Hello World.'},
            ...     {'row': 2, 'Message': 'Goodbye World.'},
            ... ])
              row  Message
            -----  --------------
                1  Hello World.
                2  Goodbye World.
        """
        from tabulate import tabulate

        if (
            obj
            and isinstance(obj, collections.abc.Sequence)
            and isinstance(obj[0], dict)
            and "headers" not in kwargs
        ):
            kwargs["headers"] = "keys"

        cls._print(tabulate(obj, **kwargs), print_kwargs=print_kwargs)

    @classmethod
    def pretty(
        cls,
        obj: Any,
        print_kwargs: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> None:
        """Pretty-print objects.

        Arguments:
            obj: object to print.
            print_kwargs: kwargs passed to `print`.
            **kwargs: kwargs passed to `pprint.pformat`.

        Examples:
            >>> import myke
            ...
            >>> myke.echo.pretty('Hello World.')
            'Hello World.'
        """
        from pprint import pformat

        cls._print(pformat(obj, **kwargs), print_kwargs=print_kwargs)

    @classmethod
    def tasks(cls, prog: Optional[str] = None, tablefmt: Optional[str] = None) -> None:
        """Print a table of registered myke tasks.

        Args:
            prog: ...
            tablefmt: table format to use (from `tabulate`)

        Examples:
            >>> import myke
            ...
            >>> @myke.task  # doctest: +SKIP
            ... def say_hello(name):
            ...     print(f"Hello {name}.")
            ...
            >>> @myke.task  # doctest: +SKIP
            ... def say_goodbye(name):
            ...     print(f"Goodbye {name}.")
            >>> myke.echo.tasks()
            <BLANKLINE>
            ===========  ==========
            Task         Source
            ===========  ==========
            say-goodbye  myke.tasks
            say-hello    myke.tasks
            ===========  ==========
            <BLANKLINE>
            To view task parameters, see:
            > myke <task-name> --help
            <BLANKLINE>
        """
        cls.text()

        if not TASKS:
            cls.text("No tasks found.")
        else:
            records: List[Dict[str, str]] = [
                {"Task": k, "Source": v.__module__} for k, v in sorted(TASKS.items())
            ]

            if not prog:
                prog = MYKE_VAR_NAME

            try:
                if not tablefmt:
                    tablefmt = "rst"
                echo.table(records, tablefmt=tablefmt)
            except ModuleNotFoundError:
                cls.text("TASKS")
                cls.text("-----")
                cls.lines([str(x) if x else "" for x in records])

            cls.text()
            cls.text("To view task parameters, see:")
            cls.text(f"> {prog} <task-name> --help")

        cls.text()



import os
from typing import List, Pattern

import mockish
import pytest
from _pytest.capture import CaptureFixture, CaptureResult

import myke
from myke.main import main
from myke.main import sys as target_sys


def test_main_version(capsys: CaptureFixture, version_pattern: Pattern):
    # 1. ARRANGE
    args: List[str] = ["--myke-version"]

    # 2. ACT
    with mockish.patch.object(target_sys, "argv", ["", *args]), pytest.raises(
        SystemExit,
    ) as e:
        main()

    # 3. ASSERT
    assert e.value.code == 0

    captured: CaptureResult = capsys.readouterr()
    assert not captured.err
    assert captured.out
    assert version_pattern.search(captured.out.split()[-1]), "invalid version"


def test_main_help(capsys: CaptureFixture):
    # 1. ARRANGE
    args: List[str] = ["--myke-help"]

    # 2. ACT
    with mockish.patch.object(target_sys, "argv", ["", *args]), pytest.raises(
        SystemExit,
    ) as e:
        main()

    # 3. ASSERT
    assert e.value.code == 0

    captured: CaptureResult = capsys.readouterr()
    assert not captured.err
    assert captured.out


def test_main_explain(capsys: CaptureFixture, resources_dir: str):
    # 1. ARRANGE
    args: List[str] = ["--myke-explain", "hello"]

    myke.TASKS.clear()
    mykefile: str = os.path.join(resources_dir, "Mykefile")

    expected: str = '''
@task\ndef hello(name="world"):\n    """Say hello."""\n    print("hello " + name)
'''

    # 2. ACT
    myke.import_mykefile(mykefile)

    with mockish.patch.object(target_sys, "argv", ["", *args]), pytest.raises(
        SystemExit,
    ) as e:
        main(mykefile)

    # 3. ASSERT
    assert e.value.code == 0

    captured: CaptureResult = capsys.readouterr()
    assert not captured.err
    assert captured.out

    assert expected.strip() == captured.out.strip()



"""> Functions for reading."""

import sys
from functools import partial, wraps
from pathlib import Path
from typing import Any, Callable, Dict, List, Union

if sys.version_info >= (3, 10):
    from typing import TypeGuard
else:
    from typing_extensions import TypeGuard


class read(str):
    def __new__(cls, path: str, encoding: str = "utf-8") -> str:  # type: ignore
        return cls.text(path=path, encoding=encoding)

    @staticmethod
    def _is_simple_dict(candidate: Any) -> TypeGuard[Dict[str, Any]]:
        return isinstance(candidate, dict) and all(
            isinstance(k, str) for k in candidate
        )

    @classmethod
    def _read_simple_dict(
        cls,
        reader: Callable[..., Any],
        **kwargs: Any,
    ) -> Dict[str, Any]:
        content: Any = reader(**kwargs)
        if not cls._is_simple_dict(content):
            raise TypeError("expected a dictionary with string keys")
        return content

    @staticmethod
    def text(path: Union[str, Path], encoding: str = "utf-8") -> str:
        """Read text file contents and strip surrounding whitespace.

        Equivalent to: `Path(path).read_text().strip()`

        Args:
            path: ...
            encoding: ...

        Returns:
            ...

        Examples:
            >>> import myke
            ...
            >>> myke.read.text('/path/to/file.txt')  # doctest: +SKIP
        """
        if isinstance(path, str):
            path = Path(path)
        return path.read_text(encoding=encoding).strip()

    @classmethod
    @wraps(text)
    def lines(cls, *args: str, **kwargs: str) -> List[str]:
        """Read lines from a text file, strip whitespace from each line, and return list of non-empty elements.

        Args:
            *args: ...
            **kwargs: ...

        Returns:
            ...

        Examples:
            >>> import myke
            ...
            >>> myke.read.lines('/path/to/file.txt')  # doctest: +SKIP
        """
        return [
            y for x in cls.text(*args, **kwargs).splitlines() for y in [x.strip()] if y
        ]

    @classmethod
    @wraps(text)
    def json(cls, *args: str, **kwargs: str) -> Dict[str, Any]:
        """Parse object(s) from a JSON text file.

        Args:
            *args: ...
            **kwargs: ...

        Returns:
            ...

        Examples:
            >>> import myke
            ...
            >>> myke.read.json('/path/to/file.json')  # doctest: +SKIP
        """
        import json as _json

        return cls._read_simple_dict(partial(_json.loads, cls.text(*args, **kwargs)))

    @classmethod
    @wraps(text)
    def yaml(cls, *args: str, **kwargs: str) -> Dict[str, Any]:
        """Parse object(s) from a YAML text file.

        Args:
            *args: ...
            **kwargs: ...

        Returns:
            ...

        Examples:
            >>> import myke
            ...
            >>> myke.read.yaml('/path/to/file.yaml')  # doctest: +SKIP
        """
        import yaml as _yaml

        return cls._read_simple_dict(
            partial(_yaml.safe_load, cls.text(*args, **kwargs)),
        )

    @classmethod
    @wraps(text)
    def yaml_all(cls, *args: str, **kwargs: str) -> List[Dict[str, Any]]:
        """Parse object(s) from multiple documents in a single YAML text file.

        Args:
            *args: ...
            **kwargs: ...

        Returns:
            ...

        Examples:
            >>> import myke
            ...
            >>> myke.read.yaml_all('/path/to/file.yaml')  # doctest: +SKIP
        """
        import yaml as _yaml

        def _yaml_all(txt: str) -> List[Dict[str, Any]]:
            return [
                cls._read_simple_dict(lambda y: y, y=x)
                for x in _yaml.safe_load_all(txt)
            ]

        return _yaml_all(cls.text(*args, **kwargs))

    @classmethod
    @wraps(text)
    def toml(cls, *args: str, **kwargs: str) -> Dict[str, Any]:
        """Parse object(s) from a TOML text file.

        Args:
            *args: ...
            **kwargs: ...

        Returns:
            ...

        Examples:
            >>> import myke
            ...
            >>> myke.read.toml('/path/to/file.toml')  # doctest: +SKIP
        """
        if sys.version_info >= (3, 11):
            import tomllib as _toml
        else:
            import tomli as _toml

        return cls._read_simple_dict(partial(_toml.loads, cls.text(*args, **kwargs)))

    @classmethod
    @wraps(text)
    def cfg(cls, *args: str, **kwargs: str) -> Dict[str, Any]:
        """Parse object(s) from a INI/CFG text file.

        Args:
            *args: ...
            **kwargs: ...

        Returns:
            ...

        Examples:
            >>> import myke
            ...
            >>> myke.read.cfg('/path/to/file.cfg')  # doctest: +SKIP
        """
        from configparser import ConfigParser

        def _read_cfg(txt: str) -> Dict[str, Any]:
            cp = ConfigParser()
            cp.read_string(txt)
            return {x: dict(cp.items(x)) for x in cp.sections()}

        return cls._read_simple_dict(partial(_read_cfg, cls.text(*args, **kwargs)))

    @classmethod
    @wraps(cfg)
    def ini(cls, *args: str, **kwargs: str) -> Dict[str, Any]:
        """Parse object(s) from a INI/CFG text file.

        Args:
            *args: ...
            **kwargs: ...

        Returns:
            ...

        Examples:
            >>> import myke
            ...
            >>> myke.read.ini('/path/to/file.ini')  # doctest: +SKIP
        """
        return cls.cfg(*args, **kwargs)

    @classmethod
    @wraps(text)
    def dotfile(cls, *args: str, **kwargs: str) -> Dict[str, str]:
        """Parse key-value pairs from a dotfile (aka "envfile").

        Args:
            *args: ...
            **kwargs: ...

        Returns:
            ...

        Examples:
            >>> import myke
            ...
            >>> myke.read.dotfile('/path/to/vars.env')  # doctest: +SKIP
        """
        from io import StringIO

        from dotenv import dotenv_values

        return cls._read_simple_dict(
            partial(dotenv_values, stream=StringIO(cls.text(*args, **kwargs))),
        )

    @classmethod
    @wraps(dotfile)
    def envfile(cls, *args: str, **kwargs: str) -> Dict[str, str]:
        """Parse key-value pairs from a dotfile (aka "envfile").

        Args:
            *args: ...
            **kwargs: ...

        Returns:
            ...

        Examples:
            >>> import myke
            ...
            >>> myke.read.envfile('/path/to/vars.env')  # doctest: +SKIP
        """
        return cls.dotfile(*args, **kwargs)

    @staticmethod
    def _url(addr: str, **kwargs: Any) -> Any:
        import requests

        addr = kwargs.pop("url", addr)
        method: str = kwargs.pop("method", "GET")
        timeout: float = kwargs.pop("timeout", 10)
        resp: requests.Response = requests.request(
            method=method,
            url=addr,
            timeout=timeout,
            **kwargs,
        )
        return resp

    @classmethod
    def url(cls, addr: str, **kwargs: Any) -> str:
        """Return text from HTTP GET response.

        Arguments:
            addr: URL of the remote file.
            **kwargs: passed to `requests.request`

        Returns:
            ...

        Examples:
            >>> import myke
            ...
            >>> myke.read.url('https://github.com/.../README.md')  # doctest: +SKIP
        """
        resp_text: str = cls._url(addr=addr, **kwargs).text
        return resp_text

    @classmethod
    def url_json(cls, addr: str, **kwargs: Any) -> Dict[str, Any]:
        """Parse JSON from HTTP GET response.

        Arguments:
            addr: URL of the remote file.
            **kwargs: passed to `requests.request`

        Returns:
            ...

        Examples:
            >>> import myke
            ...
            >>> myke.read.url_json('https://github.com/.../data.json')  # doctest: +SKIP
        """
        resp: Any = cls._url(addr=addr, **kwargs).json()
        resp_dict: Dict[str, Any] = cls._read_simple_dict(lambda: resp)
        return resp_dict
